/* 
   zmin_alpha - minimum redshift depending on matched circle size for Poincare model

   Copyright (C) 2011 Boud Roukema

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

*/

#include <stdio.h>
#include <sys/types.h>

#include <string.h>

#include "system.h"

#ifdef HAVE_ARGP_H
#  include <argp.h>
#endif  /* HAVE_ARGP_H */


#include <stdlib.h>


#define EXIT_FAILURE 1

#if ENABLE_NLS
# include <libintl.h>
# define _(Text) gettext (Text)
#else
# define textdomain(Domain)
# define _(Text) Text
#endif
#define N_(Text) Text

#include <math.h>
#include <gsl/gsl_integration.h>

#include "lib/cosmdist.h" 


char *xmalloc ();
char *xrealloc ();
char *xstrdup ();

#ifdef HAVE_ARGP_H
  static error_t parse_opt (int key, char *arg, struct argp_state *state);
  static void show_version (FILE *stream, struct argp_state *state);
#else
  static void show_version (FILE *stream);
#endif

/* argp option keys */
enum {DUMMY_KEY=129
      ,NOWARN_KEY
};

/* Option flags and variables.  These are initialized in parse_opt.  */

int want_verbose;		/* --verbose */

/* local cosmological parameters */
char *argument_string;
int  nlen_argument_string;
double H_0;
double omm;
double omrad;
double omlam;
double w_0;

char ** endptr = NULL;  /*  error handling by   strtod  -  info strtod  */
int  ch;  /* for removing end-of-line after scanf */


#ifdef HAVE_ARGP_H
static struct argp_option options[] =
{
  { "verbose",     'v',           NULL,            0,
    N_("Print more information"), 0 },
  { "H_0",     'h',           N_("HUBBLE_CONSTANT"),            0,
    N_("Hubble constant in km/s/Mpc"), 0 },
  { "omega_matter",     'm',           N_("OMEGA_MATTER"),            0,
    N_("density of non-relativistic Matter"), 0 },
  { "omega_lambda",     'l',           N_("OMEGA_LAMBDA"),            0,
    N_("Lambda - cosmological or quintessence constant"), 0 },
  { "omega_radiation",     'r',           N_("OMEGA_RADIATION"),            0,
    N_("radiation (photons)"), 0 },
  { "w_0",     'q',           N_("QUINTESSENCE_PARAMETER_0"),            0,
    N_("first order Quintessence parameter"), 0 },
  { NULL, 0, NULL, 0, NULL, 0 }
};

/* The argp functions examine these global variables.  */
const char *argp_program_bug_address = "<boud at astro.uni.torun.pl>";
void (*argp_program_version_hook) (FILE *, struct argp_state *) = show_version;

static struct argp argp =
{
  options, parse_opt, N_(" "),
  N_("Poincare dodecahedral space matched disc central redshifts: enter alpha in degrees"),
  NULL, NULL, NULL
};
#endif  /* HAVE_ARGP_H */



int
main (int argc, char **argv)
{

  /*  double H_0, omm, omlam, omrad, w_0;  */
  double input_parameter;
  double alpha=-99.9; /* input value */
  double redshift=-99.9, distance=-99.9, time=-99.9;   
  double dist=-99.9, time_output=-99.9, dist2;
  /* double z_single; */
  double red_inv, red_inv2;

  int want_debug = (0 || COSM_WANT_DEBUG);

  double omlam_1 ;
  double omlam_2 = 0.8 ;
  double del_omlam; 
  int  N_omlam = 1000; /* 100; */
  int  i_omlam;

  const double zSLS = 1100;
  const double c_kms = 299792.458; 
  double rc_direct, rSLS, rc_from_alpha;
  double omlam_best ;
  double rSLS_best, rc_best ;
  double rc_diff = 9e9; 
  double rc_diff_1, rc_diff_2;
  double rc_error;
  double min_rc_error = 9e9; /* minimum abs error between two ways of estimating R_C */
  double required_rc_error = 1.0; /* in Mpc */

  /* for testing */
  double z1log = -2.0;
  double z2log = +3.1;
  double dztest;
  double danalyt; /* analytical distance */
  double dnumer, dtest;  /* numerical distance, diff */
  double tanalyt; /* analytical time */
  double tnumer, ttest;  /* numerical time, diff */
  double zinverse, zdiff;  /* inverse redshift, difference  */
  double omm_1 = 0.1 ;
  double omm_2 = 0.9 ;
  double dommtest, omm_test ;

  double xx;  /* very local (few lines only) variable */

  double testtoler = 0.1;  /* Mpc or Gyr */
  int iztest, maxztest = 10000; 
  double testtoler_Ue_Li_Pen_less02 = 0.05;  /* relative error for omm < 0.2 */
  double testtoler_Ue_Li_Pen_more02 = 0.005;  /* relative error for omm > 0.2 */
  int maxztest_Ue_Li_Pen = 10; /* the fitting formula: only few % accuracy*/
  int iommtest, maxommtest = 9;

  textdomain(PACKAGE);

  /* Set up default values.  */
  want_verbose = 0;

  H_0 = 70.0;
  omm = 0.3;
  omlam = 0.7;
  omrad = 1.65e-4 * omm; /* TODO: derive from T_0 etc. */
  w_0 = -1.0;
  
  /*  redshift = 3.0; */

#if HAVE_ARGP_H
  argp_parse(&argp, argc, argv, 0, NULL, NULL);
#else
  printf("Warning: no command line options since a recent glibc is not installed.\n");
#endif  /* HAVE_ARGP_H */


  /* PRELIMINARY work */

  /*  redshift = 3.0;  */

  /* read parameter from standard input */

  if(want_verbose)
    {
      printf("Enter ");
      {
        printf("matched circle radius alpha in deg: ");
      };
    };

  if(want_debug)
    {
      printf("debug1: errno = %d\n",argument_string,errno);
    };
  
  /* TODO: this bit gives errno as of gfortran/lenny - should be cleaned up */
  argument_string= xmalloc(40);  /* unlikely to need more than 40 bytes */
  scanf("%40s", argument_string);
  while ( (ch=fgetc(stdin)) != '\n' && ch != EOF )
    {
    };

  if(want_debug)
    {
      printf("debug2: string= _%s_  errno = %d\n",argument_string,errno);
    };
  
  errno=0;
  input_parameter= strtod(argument_string,endptr);

  if(errno != 0)
    {
      printf("ERROR: errno= %d  in reading standard input\n",errno);
      printf("argument_string = %s\n",argument_string); /* DEBUG ONLY */
      exit (errno);
    };
  
  if(want_verbose)
    {
    printf("read parameter: %.5f \n",input_parameter);
    };

  
  alpha = input_parameter;


  /*  cosm_dist_init();  This is called by the functions if needed */

  /*
    H_0 = 100.0;
    omm = 1.0;
    omlam = 0.0;
  */

  if(want_verbose)
    {
      printf("H_0= %.2f  Omega_matter= %.5f  Omega_lambda= %.5f  Omega_radiation = %.8f",
	     H_0,omm,omlam,omrad);
      printf(" w_0= %.5f, ",w_0);
      printf(" alpha= %.2f\n",alpha);

    };


  /* 
     Using cosmdist as a C library
     This comment is referred to by cosmdist.texinfo - please do not
     remove it unless you fix both simultaneously.
  */


  /*  REAL WORK
      default: find best omlam given omm, etc., alpha 
      See Eq.~(15) \citep{RBSG08}  \tan \frac{\rSLS}{R_C} = \frac{\tan (\pi/10)}{\cos \alpha}
  */

  /* del_omlam = (omlam_2-omlam_1)/(double)N_omlam; */
  omlam_1 = 1.0 - omm + 1e-4;
  
  /*  for(i_omlam=0;i_omlam<N_omlam;i_omlam++){ */
  i_omlam=0;
  omlam=9e9;
  while(i_omlam<N_omlam && min_rc_error > required_rc_error){
    /* omlam = omlam_1 + ((double)(i_omlam)+0.5)*del_omlam; */
    switch(i_omlam){
    case 0: 
      omlam = omlam_1;
      if(want_debug)printf("AA.0: %d %.3f %.3f\n",i_omlam, omm , omlam);
      break;
    case 1:
      omlam = omlam_2;
      if(want_debug)printf("AA.1: %d %.3f %.3f\n",i_omlam, omm , omlam);
      break;
    default:
      omlam = 0.5 * (omlam_1 + omlam_2);
      if(want_debug)printf("AA.other: %d %.3f %.3f\n",i_omlam, omm , omlam);
    };

    if(want_debug)printf("AA: %d %.3f %.3f\n",i_omlam, omm , omlam);
    if(omm + omlam -1.0 > 0.0){
      rc_direct= c_kms /H_0 / sqrt(omm + omlam -1.0); /* use metric parameters directly */
      
      errno=0;
      rSLS = cosm_get_comov_dist( H_0, omm, omlam,omrad, w_0, zSLS );
      if(errno != 0){
        perror("ERROR during cosm_get_comov_dist");
      }else{
        rc_from_alpha = rSLS / atan( tan(M_PI/10.0)/cos(alpha*M_PI/180.0) ) ;
        rc_diff = rc_direct - rc_from_alpha ;
        rc_error = fabs( rc_diff );
        if(want_debug) printf(" new: %.1f %.1f %.1f, %.5f %.5f %.5f ; ",
                              rc_diff_1, rc_diff, rc_diff_2, omlam_1, omlam, omlam_2);
        if(rc_error < min_rc_error ){
          /* update best estimate parameters */
          min_rc_error  = rc_error ;  
          omlam_best = omlam;
          rSLS_best = rSLS ;
          rc_best = 0.5* (rc_direct+rc_from_alpha);
          if(want_debug) printf("better: ");
        } /* if(fabs( rc_direct - rc_from_alpha ) < min_rc_error ) */
        else {
          if(want_debug) printf("worse:  ");
        };

          /* tighten domain except at first two steps */
          if(i_omlam >= 2){ /* iterative improvement */
            if( (rc_diff / rc_diff_1) > 0.0){
              omlam_1 = omlam;
              rc_diff_1 = rc_diff;
            }else {
              omlam_2 = omlam;
              rc_diff_2 = rc_diff;
            };
          } else { /* initialise diffs if at first two steps */
              if(i_omlam==0){
                rc_diff_1 = rc_diff;
              }else{
                rc_diff_2 = rc_diff;
              };
          };

        if(want_debug){
          printf("i= %d  omm= %.5f  omlam= %.5f ",i_omlam,omm,omlam);
          printf("rc_direct= %.2f  rc_from_alpha= %.2f  rc_error= %.2f rSLS= %.2f\n",
                 rc_direct, rc_from_alpha, rc_error,rSLS );
        };
      }; /* if distance found without error */
      if(want_debug) printf(" DDD %d\n",i_omlam);
    } /* if(omm + omlam -1.0 > 0.0) */
    else { /* in this case, need to increase omlam */
      omlam_1 = omlam;
      rc_diff_1 = rc_diff;
    };
    i_omlam++;
    if(want_debug)printf("\n");
  }; /*  while( )  || for(i_omlam=0;i_omlam<N_omlam;i_omlam++) */

  if(want_debug)
    printf("done: %d %d  %.1f %.1f\n",
           i_omlam, N_omlam, min_rc_error, required_rc_error);

  if(want_verbose)
    printf("omlam_best is %.5f  rSLS_best is %.5f | omm = %.3f omtot = %.5f\n", 
           omlam_best,rSLS_best, omm, omm+omlam_best+omrad);

  /* now evaluate pi/10 R_C  and get the corresponding redshift to the matched disc centre */
  errno=0;
  zinverse= cosm_get_comov_dist_inv(H_0, omm, omlam_best, omrad, w_0, 
                                    (M_PI/10.0 * rc_best)
                                    ); 
  if(errno != 0)
    perror("ERROR during cosm_get_comov_dist_inv");
  
  if(want_verbose)
    printf("z_min is %.5f\n", zinverse);
  else
    printf("%.5f\n",zinverse);

  exit (0);


};

#ifdef HAVE_ARGP_H
/* Parse a single option.  */
static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  switch (key)
    {
    case ARGP_KEY_INIT:
      break;
    case 'v':			/* --verbose */
      want_verbose = 1;
      break;
    case 'h':
      argument_string= xstrdup(arg);
      nlen_argument_string= strlen(argument_string);
      errno=0;
      H_0= strtod(argument_string,endptr);
      if(errno != 0)
	{
	  printf("ERROR: errno= %d  in reading H_0 value\n",errno);
	  exit (errno);
	};
      break;
    case 'm':
      argument_string= xstrdup(arg);
      nlen_argument_string= strlen(argument_string);
      errno=0;
      omm= strtod(argument_string,endptr);
      if(errno != 0)
	{
	  printf("ERROR: errno= %d  in reading Omega_matter value\n",errno);
	  exit (errno);
	};
      break;
    case 'l':
      argument_string= xstrdup(arg);
      nlen_argument_string= strlen(argument_string);
      errno=0;
      omlam= strtod(argument_string,endptr);
      if(errno != 0)
	{
	  printf("ERROR: errno= %d  in reading Omega_Lambda value\n",errno);
	  exit (errno);
	};
      break;
    case 'r':
      argument_string= xstrdup(arg);
      nlen_argument_string= strlen(argument_string);
      errno=0;
      omrad= strtod(argument_string,endptr);
      if(errno != 0)
	{
	  printf("ERROR: errno= %d  in reading Omega_radiation value\n",errno);
	  exit (errno);
	};
      break;
    case 'q':
      argument_string= xstrdup(arg);
      nlen_argument_string= strlen(argument_string);
      errno=0;
      w_0= strtod(argument_string,endptr);
      if(errno != 0)
	{
	  printf("ERROR: errno= %d  in reading w_0 value\n",errno);
	  exit (errno);
	};
      break;
    case ARGP_KEY_ARG:		/* [FILE]... */
      /* TODO: Do something with ARG, or remove this case and make
         main give argp_parse a non-NULL fifth argument.  */
      break;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}
#endif  /* HAVE_ARGP_H */


/* Show the version number and copyright information.  */
static void
#ifdef HAVE_ARGP_H
show_version (FILE *stream, struct argp_state *state)
#else
     show_version (FILE *stream)
#endif  /* HAVE_ARGP_H */
{
#ifdef HAVE_ARGP_H
  (void) state;
#endif  /* HAVE_ARGP_H */
  /* Print in small parts whose localizations can hopefully be copied
     from other programs.  */
  fputs(PACKAGE" "VERSION"\n", stream);
  fprintf(stream, _("Written by %s.\n\n"), "Boud Roukema");
  fprintf(stream, _("Copyright (C) %s %s\n"), "2004", "Boud Roukema");
  fputs(_("\
This program is free software; you may redistribute it under the terms of\n\
the GNU General Public License.  This program has absolutely no warranty.\n"),
	stream);
}
